/*
    ball-pack.cpp - a collection of distance balls

    CC BY-SA Edouard.Thiel@univ-amu.fr - 04/08/2023
*/

#include "ball-pack.hpp"


BallPack::BallPack()
{
    ball_d4.add_chamfer_weighting (1, 0, 1);

    ball_d8.add_chamfer_weighting (1, 0, 1);
    ball_d8.add_chamfer_weighting (1, 1, 1);

    ball_d23.add_chamfer_weighting (1, 0, 2);
    ball_d23.add_chamfer_weighting (1, 1, 3);

    ball_d34.add_chamfer_weighting (1, 0, 3);
    ball_d34.add_chamfer_weighting (1, 1, 4);

    ball_d5711.add_chamfer_weighting (1, 0, 5);
    ball_d5711.add_chamfer_weighting (1, 1, 7);
    ball_d5711.add_chamfer_weighting (2, 1, 11);
}


std::string BallPack::get_dist_id_by_name (DistanceId dist_id)
{
    switch (dist_id) {
        case M_D4      : return "d4";
        case M_D8      : return "d8";
        case M_2_3     : return "d_2_3";
        case M_3_4     : return "d_3_4";
        case M_5_7_11  : return "d_5_7_11";
        case M_DE2     : return "dE2";
        case M_DPF1    : return "dpF1";
        case M_DPF0    : return "dpF0";
        default        : return "unknown";
    }
}


BallPack::DistanceId BallPack::get_dist_name_by_id (std::string name)
{
    if (name == "d4") return M_D4;
    if (name == "d8") return M_D8;
    if (name == "d_2_3") return M_2_3;
    if (name == "d_3_4") return M_3_4;
    if (name == "d_5_7_11") return M_5_7_11;
    if (name == "dE2") return M_DE2;
    if (name == "dpF1") return M_DPF1;
    if (name == "dpF0") return M_DPF0;
    return M_LAST;
}


std::string BallPack::get_available_dist_names ()
{
    std::string s;
    
    for (auto d = M_D4 ; d < M_LAST ; 
         d = static_cast<DistanceId>(static_cast<int>(d)+1) )
    {
        if (d > M_D4) s+= ", ";
        s += "'"+get_dist_id_by_name(d)+"'";
    }
    
    return s;
}


BallPack::DistanceId BallPack::get_prev_dist_id (DistanceId dist_id)
{
    int i = static_cast<int>(dist_id);
    if (i == 0) i = static_cast<int>(M_LAST);
    return static_cast<DistanceId>(i-1);
}


BallPack::DistanceId BallPack::get_next_dist_id (DistanceId dist_id)
{
    DistanceId id_next = static_cast<DistanceId>(static_cast<int>(dist_id)+1);
    if (id_next == M_LAST) id_next = M_D4;
    return id_next;
}


std::string BallPack::get_dist_name ()
{
    return get_dist_id_by_name (current_dist_id);
}


void BallPack::set_prev_dist ()
{
    current_dist_id = get_prev_dist_id (current_dist_id);
}


void BallPack::set_next_dist ()
{
    current_dist_id = get_next_dist_id (current_dist_id);
}


void BallPack::compute_mlutg (int L, int opt_dt)
{
    switch (current_dist_id) {
        case M_D4      : ball_d4.compute_mlutg (L, opt_dt); break;
        case M_D8      : ball_d8.compute_mlutg (L, opt_dt); break;
        case M_2_3     : ball_d23.compute_mlutg (L, opt_dt); break;
        case M_3_4     : ball_d34.compute_mlutg (L, opt_dt); break;
        case M_5_7_11  : ball_d5711.compute_mlutg (L, opt_dt); break;
        case M_DE2     : ball_dE2.compute_mlutg (L, opt_dt); break;
        case M_DPF0    : ball_dpF0.compute_mlutg (L, opt_dt); break;
        case M_DPF1    : ball_dpF1.compute_mlutg (L, opt_dt); break;
        default        : ;
    }
}


void BallPack::compute_DT (cv::Mat &img_int)
{
    switch (current_dist_id) {
        case M_D4      : ball_d4.compute_WDT(img_int); break;
        case M_D8      : ball_d8.compute_WDT(img_int); break;
        case M_2_3     : ball_d23.compute_WDT(img_int); break;
        case M_3_4     : ball_d34.compute_WDT(img_int); break;
        case M_5_7_11  : ball_d5711.compute_WDT(img_int); break;
        case M_DE2     : ball_dE2.compute_SEDT_Saito_Toriwaki(img_int); break;
        case M_DPF0    : ball_dpF0.compute_SEDT_Saito_Toriwaki(img_int); break;
        case M_DPF1    : ball_dpF1.compute_SEDT_Saito_Toriwaki(img_int); break;
        default        : ;
    }
}


void BallPack::compute_MA (cv::Mat &img_int)
{
    switch (current_dist_id) {
        case M_D4      : ball_d4.compute_medial_axis(img_int); break;
        case M_D8      : ball_d8.compute_medial_axis(img_int); break;
        case M_2_3     : ball_d23.compute_medial_axis(img_int); break;
        case M_3_4     : ball_d34.compute_medial_axis(img_int); break;
        case M_5_7_11  : ball_d5711.compute_medial_axis(img_int); break;
        case M_DE2     : ball_dE2.compute_medial_axis(img_int); break;
        case M_DPF0    : ball_dpF0.compute_medial_axis(img_int); break;
        case M_DPF1    : ball_dpF1.compute_medial_axis(img_int); break;
        default        : ;
    }
}


void BallPack::compute_RDT (cv::Mat &img_int)
{
    switch (current_dist_id) {
        case M_D4      : ball_d4.compute_RWDT(img_int); break;
        case M_D8      : ball_d8.compute_RWDT(img_int); break;
        case M_2_3     : ball_d23.compute_RWDT(img_int); break;
        case M_3_4     : ball_d34.compute_RWDT(img_int); break;
        case M_5_7_11  : ball_d5711.compute_RWDT(img_int); break;
        case M_DE2     : ball_dE2.compute_RSEDT_Saito_Toriwaki(img_int); break;
        case M_DPF0    : ball_dpF0.compute_RSEDT_Saito_Toriwaki(img_int); break;
        case M_DPF1    : ball_dpF1.compute_RSEDT_Saito_Toriwaki(img_int); break;
        default        : ;
    }
}


void BallPack::compute_level_curves (cv::Mat &img_int)
{
    switch (current_dist_id) {
        case M_D4      : ball_d4.compute_level_curves(img_int); break;
        case M_D8      : ball_d8.compute_level_curves(img_int); break;
        case M_2_3     : ball_d23.compute_level_curves(img_int); break;
        case M_3_4     : ball_d34.compute_level_curves(img_int); break;
        case M_5_7_11  : ball_d5711.compute_level_curves(img_int); break;
        case M_DE2     : ball_dE2.compute_level_curves(img_int); break;
        case M_DPF0    : ball_dpF0.compute_level_curves(img_int); break;
        case M_DPF1    : ball_dpF1.compute_level_curves(img_int); break;
        default        : ;
    }
}


