/*
    ball.cpp - Mlut and LUT columns computation for a ball family in 2D

    Original papers and sources:
      - E. Remy and E. Thiel. Exact Medial Axis with Euclidean Distance. 
        Image and Vision Computing, 23(2):167-175, 2005. 
        https://pageperso.lis-lab.fr/edouard.thiel/IVC2004/index.html

      - E. Remy and E. Thiel. Medial Axis for Chamfer Distances: computing 
        Look-Up Tables and Neighbourhoods in 2D or 3D. 
        Pattern Recognition Letters, 23(6):649-661, April 2002.
        https://pageperso.lis-lab.fr/edouard.thiel/PRL2001/index.html

    CC BY-SA Edouard.Thiel@univ-amu.fr, Eric.Remy@univ-amu.fr - 04/08/2023
*/

#include "mask-lut.hpp"


/*
 * Create a new image L*L and initialize the pixels to 0.
 *
 * Input:  L the side length.
 * Output: the pointer to the image, or NULL if memory allocation failed.
*/

int *create_new_image (int L)
{
    int n = L*L;
    int *res;

    //printf ("create_new_image: allocating %ldk ... ", n*sizeof(int)/1024);
    res = (int*) calloc (n, sizeof(int));
    if (res == NULL)
         printf ("malloc error\n");
    return res;
}


void print_image (int L, int *img)
{
    for (int y = 0; y < L; y++) {
        for (int x = 0; x < L; x++) {
            if (x < y) printf ("     ");
            else printf ("%4d ", (int)img[ y*L + x ]);
        }
        printf("\n");
    }
}


void MaskLut::reset()
{
    mlutg.clear();
    mlut_full.clear();
    possible.clear();
    Rknown = 0;
}


/*
 * Add a weighting in the mask mlutg.
 *
 * Input:  (x,y,r) the weighting.
 * Output: the number of the weighting in the mask.
*/

int MaskLut::add_mlutg_weighting (int x, int y, int r)
{
    if (! (0 <= y && y <= x && 0 < x && 0 < r)) {
        printf ("add_mlutg_weighting: (x = %d, y = %d, r = %d)\n", x, y, r);
        printf ("  does not respect 0 <= y <= x, 0 < x, 0 < r\n");
        return 0;
    }
    MLutG_weighting vg;
    vg.x = x;
    vg.y = y;
    vg.r = r;
    mlutg.push_back(vg);
    return SIZE_INT(mlutg)-1;
}


void MaskLut::add_mlut_weigthing (int x, int y, int ind_vg)
{
    MLut_weighting v;
    v.x = x;
    v.y = y;
    v.ind_vg = ind_vg;
    mlut_full.push_back(v);
}


void MaskLut::populate_mlut_full()
{
    mlut_full.clear();
    for (int i = 0; i < SIZE_INT(mlutg); i++)
    {
        MLutG_weighting &vg = mlutg[i];
        if (vg.y == 0) {
            add_mlut_weigthing ( vg.x, vg.y, i);
            add_mlut_weigthing (-vg.x, vg.y, i);
            add_mlut_weigthing ( vg.y, vg.x, i);
            add_mlut_weigthing ( vg.y,-vg.x, i);
        } else if (vg.x == vg.y) {
            add_mlut_weigthing ( vg.x, vg.y, i);
            add_mlut_weigthing (-vg.x, vg.y, i);
            add_mlut_weigthing ( vg.x,-vg.y, i);
            add_mlut_weigthing (-vg.x,-vg.y, i);
       } else {
            add_mlut_weigthing ( vg.x, vg.y, i);
            add_mlut_weigthing ( vg.y, vg.x, i);
            add_mlut_weigthing (-vg.x, vg.y, i);
            add_mlut_weigthing (-vg.y, vg.x, i);
            add_mlut_weigthing ( vg.x,-vg.y, i);
            add_mlut_weigthing ( vg.y,-vg.x, i);
            add_mlut_weigthing (-vg.x,-vg.y, i);
            add_mlut_weigthing (-vg.y,-vg.x, i);
       }
    }
}


/* Miscellaneous */

int compute_gcd (int a, int b)
{
    if (a*b == 0) return a+b;
    if (a <= b) return compute_gcd (a, b % a);
    return compute_gcd (a % b, b);
}

int is_visible (int x, int y)
{
    return compute_gcd(x,y) == 1;
}


/*
 * Lut column Computation Algorithm.
 *
 * Input:  CTg the distance cone to origin, L the side length, vg the
 *         weighting, Rmax the greatest radius
 *         to be verified in Lut.
 * Output: The Lut column vg.lut_col is filled with the correct values.
*/

void MaskLut::compute_lut_col (int *CTg, int L, MLutG_weighting &vg, int Rmax)
{
    int x, y, r1, r2, ra, rb;

    /* Initializes the Lut to 0 */
    vg.lut_col.assign (Rmax+1, 0);

    for (x = 0; x < L - vg.x; x++)
    for (y = 0; y <= x; y++)
    {
        /* Radius of the ball where p1 is located */
        r1 = CTg[ y*L + x ] +1;

        /* Same for p2 */
        r2 = CTg[ (y+vg.y)*L + x+vg.x ] +1;

        if (r1 <= Rmax && r2 > vg.lut_col[r1]) vg.lut_col[r1] = r2;
    }

    rb = 0;
    for (ra = 0; ra <= Rmax; ra++)
    {
        if (vg.lut_col[ra] > rb)
             rb = vg.lut_col[ra];
        else vg.lut_col[ra] = rb;
    }
}


/*
 * Fast extraction of MA points from Bd inter G(Z^2).
 *
 * Input:  x,y the point to test, DTg the distance transform of the section of
 *         the ball, L the side length.
 * Output: returns 1 if point x,y is detected as MA in the DTg.
*/

int MaskLut::is_MAg (int x, int y, int *DTg, int L)
{
    int xx, yy, val, i;

    val = DTg[ y*L + x ];

    for (i = 0; i < SIZE_INT(mlutg); i++)
    {
        MLutG_weighting &vg = mlutg[i];
        xx = x - vg.x;
        yy = y - vg.y;

        if (0 <= yy && yy <= xx)
        {
            if ( DTg[ yy*L + xx ] >= vg.lut_col[val] )
                return 0;
        }
    }

    return 1;
}


void MaskLut::compute_possible_values (int L, int *CTg, int rmax)
{
    possible.assign (rmax+1, false);

    for (int x = 1; x < L; x++)
    for (int y = 0; y <= x; y++)
    {
        int val = CTg[ y*L + x ];
        if (val <= rmax) possible[val] = true;
    }
}


bool MaskLut::is_possible (int R)
{
    return (R < 0 || R >= SIZE_INT(possible)) ? false : possible[R];
}


/*
 * Full Lut Computation Algorithm with determination of MgLut.
 *
 * Input:  CTg and DTg two images, L the side length, 
 *         Rtarget the maximum radius to be verified.
 * Output: mlutg is filled with the correct values.
*/

void MaskLut::compute_lut_mask (int *CTg, int *DTg, int L, int Rtarget, int opt_dt)
{
    int x, y, R, i;

    printf("compute_lut_mask for L = %d, Rtarget = %d\n", L, Rtarget);

    compute_CTg (L, CTg);
    compute_possible_values (L, CTg, Rtarget);

    if (opt_dt) {
        printf ("Image CTg :\n");
        print_image (L, CTg);
    }

    /* Compute Lut columns for current Lut mask */
    for (i = 0; i < SIZE_INT(mlutg); i++)
        compute_lut_col (CTg, L, mlutg[i], Rtarget);

    for (R = Rknown+1; R <= Rtarget; R++)
    if (is_possible(R))
    {
        //printf ("R = %d / %d\r", R, Rtarget); fflush (stdout);

        /* Here we can avoid to init DTg because the ball grows with R. */
        compute_DTg (L, CTg, DTg, R);

        if (opt_dt) {
            printf ("Image DTg : for R = %d\n", R);
            print_image (L, DTg);
        }

        for (x = 1; x < L; x++)
        {
            if (DTg[0*L + x] == 0) break;  /* Outside the ball */
            
            for (y = 0; y <= x; y++)
            {
                if (DTg[y*L + x] == 0) break; /* Outside the ball */
                
                if ( is_MAg (x, y, DTg, L) )
                {
                    /* Add a new weighting to MgL */
                    i = add_mlutg_weighting (x, y, R);
    
                    printf ("i=%3d  (x,y)= ( %2d, %2d)  added for R= %7d   %s\n",
                        i+1, x, y, R,
                        is_visible(x,y)?"visible":"** NON VISIBLE");
                    fflush(stdout);
    
                    /* New column in Lut */
                    compute_lut_col (CTg, L, mlutg[i], Rtarget);
    
                    if (is_MAg (x, y, DTg, L))
                    {
                        printf ("\ncompute_lut_mask: ERROR for R = %d\n", R);
                        return;
                    }
                }
            }
        }
    }

    if (Rtarget > Rknown) Rknown = Rtarget;
}


/*
 * Print mlutg in a file.
 *
 * Input:  f a file
 * Output: to file f.
*/

void MaskLut::print_mlutg (FILE *f)
{
    int i;

    fprintf (f, "# Computed Lut Mask:\n");
    fprintf (f, "# i  (  x,  y)        R\n");
    for (i = 0; i < SIZE_INT(mlutg); i++)
        fprintf (f, "%3d  ( %2d, %2d)  %7d\n",
            i+1, mlutg[i].x, mlutg[i].y, mlutg[i].r);
    fprintf (f, "\n");
}


/*
 * Print the LUT columns in a file.
 *
 * Input:  f a file, CTg the distance cone to origin, L the side length.
 * Output: print the lut colums to file f.
*/

void MaskLut::print_lut_cols (FILE *f, int L)
{
    int i, j;

    fprintf (f, "# Look-Up Table: (L = %d, R <= %d)\n", L, Rknown);

    for (j = 1; j <= Rknown; j++)
    if (is_possible(j))
    {
        fprintf (f, "Lut[][%5d ] = ", j);

        for (i = 0; i < SIZE_INT(mlutg); i++)
            fprintf (f, "%7d ", mlutg[i].lut_col[j]);

        fprintf (f, "\n");
    }
    fprintf (f, "\n");
}


/*
 * Computation of MLutG for image size L
*/

void MaskLut::compute_mlutg (int L, int opt_dt)
{
    int *CTg = create_new_image (L);
    int *DTg = create_new_image (L);
    int Rtarget = get_greatest_radius (L);
    compute_lut_mask (CTg, DTg, L, Rtarget, opt_dt);

    free (CTg); free (DTg);
}


/*
 * Computation of MLutG for the maximum DT value Rmax
*/

void MaskLut::compute_mlutg_for_Rmax (int L, int Rmax, int opt_dt)
{
    int *CTg = create_new_image (L);
    int *DTg = create_new_image (L);

    compute_lut_mask (CTg, DTg, L, Rmax, opt_dt);

    free (CTg); free (DTg);
}


//----------------------------- PART USING OPENCV -----------------------------


int MaskLut::get_img_max_value(cv::Mat &img)
{
    CHECK_MAT_TYPE(img, CV_32SC1);

    int vmax = img.at<int>(0, 0);

    for (int y = 0; y < img.rows; y++)
    for (int x = 0; x < img.cols; x++)
    {
        int v = img.at<int>(y,x);
        if (v > vmax) vmax = v;
    }

    return vmax;
}


bool MaskLut::is_medial_axis_point (cv::Mat &img_dt, int x, int y)
{
    CHECK_MAT_TYPE(img_dt, CV_32SC1);

    int val = img_dt.at<int>(y,x);
    if (val <= 0 || val > Rknown) return false;

    for (int i = 0; i < SIZE_INT(mlut_full); i++) {
        MLut_weighting &v = mlut_full[i];
        std::vector<int> &lut_col = mlutg[v.ind_vg].lut_col;

        int xa = x + v.x, ya = y + v.y; 
        if ( xa >= 0 && xa < img_dt.cols &&
             ya >= 0 && ya < img_dt.rows &&
             img_dt.at<int>(ya,xa) >= lut_col[val] )
          return false;
    }

    return true;
}


void MaskLut::compute_medial_axis (cv::Mat &img)
{
    CHECK_MAT_TYPE(img, CV_32SC1);

    int L = std::max (img.rows, img.cols);
    int Rmax = get_img_max_value(img);
    compute_mlutg_for_Rmax (L, Rmax, 0);
    populate_mlut_full();

    // We need to keep a copy
    cv::Mat img_dt = img.clone();

    for (int y = 0; y < img_dt.rows; y++)
    for (int x = 0; x < img_dt.cols; x++)
    {
        img.at<int>(y,x) = is_medial_axis_point (img_dt, x, y) ?
            img_dt.at<int>(y,x) : 0;
    }
}


