/*
    mask-lut.hpp - Mlut and LUT columns computation for a ball family in 2D

    Original papers and sources:
      - E. Remy and E. Thiel. Exact Medial Axis with Euclidean Distance. 
        Image and Vision Computing, 23(2):167-175, 2005. 
        https://pageperso.lis-lab.fr/edouard.thiel/IVC2004/index.html

      - E. Remy and E. Thiel. Medial Axis for Chamfer Distances: computing 
        Look-Up Tables and Neighbourhoods in 2D or 3D. 
        Pattern Recognition Letters, 23(6):649-661, April 2002.
        https://pageperso.lis-lab.fr/edouard.thiel/PRL2001/index.html

    CC BY-SA Edouard.Thiel@univ-amu.fr, Eric.Remy@univ-amu.fr - 04/08/2023
*/

#ifndef MASK_LUT_H
#define MASK_LUT_H

#include <cstdio>
#include <iostream>
#include <vector>
#include "astico2d.hpp"


#define SIZE_INT(items)  static_cast<int>(items.size())


struct MLutG_weighting {
    int x, y;                   // coordinates in G(Z²)
    int r;                      // appearence radius
    std::vector<int> lut_col;   // LUT column for this weighting
};


struct MLut_weighting {
    int x, y;                   // coordinates in Z²
    int ind_vg;                 // index in mlutg
};


class MaskLut {

  public :
    std::vector<MLutG_weighting> mlutg;
    std::vector<MLut_weighting> mlut_full;
    std::vector<bool>possible;
    int Rknown = 0;                     // the maximum verified radius

    virtual void compute_CTg (int L, int *CTg) = 0;
    virtual void compute_DTg (int L, int *CTg, int *DTg, int R) = 0;
    virtual int get_greatest_radius (int L) = 0;

    void reset();
    int add_mlutg_weighting (int x, int y, int r);
    void add_mlut_weigthing (int x, int y, int ind_vg);
    void populate_mlut_full();
    void compute_possible_values (int L, int *CTg, int rmax);
    bool is_possible (int R);
    void compute_lut_col (int *CTg, int L, MLutG_weighting &vg, int Rmax);
    int is_MAg (int x, int y, int *DTg, int L);
    void compute_lut_mask (int *CTg, int *DTg, int L, int Rtarget, 
        int opt_dt);
    void print_mlutg (FILE *f);
    void print_lut_cols (FILE *f, int L);
    void compute_mlutg (int L, int opt_dt);
    void compute_mlutg_for_Rmax (int L, int Rmax, int opt_dt);

    int get_img_max_value(cv::Mat &img);
    void populate_full_mlut ();
    bool is_medial_axis_point (cv::Mat &img_dt, int x, int y);
    void compute_medial_axis (cv::Mat &img);

};  // class MaskLut


int *create_new_image (int L);
void print_image (int L, int *img);
int compute_gcd (int a, int b);
int is_visible (int x, int y);


#endif // MASK_LUT_H

